@@ -1,26 +1,6 @@
-/* Spa
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Spa */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <errno.h>
 #include <string.h>
@@ -41,13 +21,17 @@
 
 #include "mix-ops.h"
 
-#define NAME "audiomixer"
+#undef SPA_LOG_TOPIC_DEFAULT
+#define SPA_LOG_TOPIC_DEFAULT log_topic
+static struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.audiomixer");
+
+#define DEFAULT_RATE		48000
+#define DEFAULT_CHANNELS	2
 
-#define MAX_SAMPLES     8192
 #define MAX_BUFFERS     64
 #define MAX_PORTS       128
 #define MAX_CHANNELS    64
-#define MAX_BUFFER_SIZE    (MAX_SAMPLES * MAX_CHANNELS * 8)
+#define MAX_ALIGN	MIX_OPS_MAX_ALIGN
 
 #define PORT_DEFAULT_VOLUME	1.0
 #define PORT_DEFAULT_MUTE	false
@@ -72,8 +56,6 @@ struct buffer {
 	struct spa_buffer *buffer;
 	struct spa_meta_header *h;
 	struct spa_buffer buf;
-	struct spa_data datas[1];
-	struct spa_chunk chunk[1];
 };
 
 struct port {
@@ -105,6 +87,8 @@ struct impl {
 	struct spa_log *log;
 	struct spa_cpu *cpu;
 	uint32_t cpu_flags;
+	uint32_t max_align;
+	uint32_t quantum_limit;
 
 	struct mix_ops ops;
 
@@ -116,7 +100,7 @@ struct impl {
 
 	uint32_t port_count;
 	uint32_t last_port;
-	struct port in_ports[MAX_PORTS];
+	struct port *in_ports[MAX_PORTS];
 	struct port out_ports[1];
 
 	int n_formats;
@@ -126,16 +110,14 @@ struct impl {
 	unsigned int started:1;
 	uint32_t stride;
 	uint32_t blocks;
-
-	uint8_t empty[MAX_BUFFER_SIZE];
-
 };
 
-#define CHECK_FREE_IN_PORT(this,d,p) ((d) == SPA_DIRECTION_INPUT && (p) < MAX_PORTS && !this->in_ports[(p)].valid)
-#define CHECK_IN_PORT(this,d,p)      ((d) == SPA_DIRECTION_INPUT && (p) < MAX_PORTS && this->in_ports[(p)].valid)
+#define PORT_VALID(p)                ((p) != NULL && (p)->valid)
+#define CHECK_FREE_IN_PORT(this,d,p) ((d) == SPA_DIRECTION_INPUT && (p) < MAX_PORTS && !PORT_VALID(this->in_ports[(p)]))
+#define CHECK_IN_PORT(this,d,p)      ((d) == SPA_DIRECTION_INPUT && (p) < MAX_PORTS && PORT_VALID(this->in_ports[(p)]))
 #define CHECK_OUT_PORT(this,d,p)     ((d) == SPA_DIRECTION_OUTPUT && (p) == 0)
 #define CHECK_PORT(this,d,p)         (CHECK_OUT_PORT(this,d,p) || CHECK_IN_PORT (this,d,p))
-#define GET_IN_PORT(this,p)          (&this->in_ports[p])
+#define GET_IN_PORT(this,p)          (this->in_ports[p])
 #define GET_OUT_PORT(this,p)         (&this->out_ports[p])
 #define GET_PORT(this,d,p)           (d == SPA_DIRECTION_INPUT ? GET_IN_PORT(this,p) : GET_OUT_PORT(this,p))
 
@@ -216,7 +198,7 @@ static int impl_node_add_listener(void *object,
 	emit_node_info(this, true);
 	emit_port_info(this, GET_OUT_PORT(this, 0), true);
 	for (i = 0; i < this->last_port; i++) {
-		if (this->in_ports[i].valid)
+		if (PORT_VALID(this->in_ports[i]))
 			emit_port_info(this, GET_IN_PORT(this, i), true);
 	}
 
@@ -243,6 +225,12 @@ static int impl_node_add_port(void *object, enum spa_direction direction, uint32
 	spa_return_val_if_fail(CHECK_FREE_IN_PORT(this, direction, port_id), -EINVAL);
 
 	port = GET_IN_PORT(this, port_id);
+	if (port == NULL) {
+		port = calloc(1, sizeof(struct port));
+		if (port == NULL)
+			return -errno;
+		this->in_ports[port_id] = port;
+	}
 	port->direction = SPA_DIRECTION_INPUT;
 	port->id = port_id;
 
@@ -269,7 +257,7 @@ static int impl_node_add_port(void *object, enum spa_direction direction, uint32
 		this->last_port = port_id + 1;
 	port->valid = true;
 
-	spa_log_debug(this->log, NAME " %p: add port %d:%d %d", this,
+	spa_log_debug(this->log, "%p: add port %d:%d %d", this,
 			direction, port_id, this->last_port);
 	emit_port_info(this, port, true);
 
@@ -299,12 +287,12 @@ impl_node_remove_port(void *object, enum spa_direction direction, uint32_t port_
 		int i;
 
 		for (i = this->last_port - 1; i >= 0; i--)
-			if (GET_IN_PORT (this, i)->valid)
+			if (PORT_VALID(GET_IN_PORT(this, i)))
 				break;
 
 		this->last_port = i + 1;
 	}
-	spa_log_debug(this->log, NAME " %p: remove port %d:%d %d", this,
+	spa_log_debug(this->log, "%p: remove port %d:%d %d", this,
 			direction, port_id, this->last_port);
 
 	spa_node_emit_port_info(&this->hooks, direction, port_id, NULL);
@@ -348,8 +336,10 @@ static int port_enum_formats(void *object,
 								SPA_AUDIO_FORMAT_U24_32,
 								SPA_AUDIO_FORMAT_F32,
 								SPA_AUDIO_FORMAT_F64),
-				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(44100, 1, INT32_MAX),
-				SPA_FORMAT_AUDIO_channels, SPA_POD_CHOICE_RANGE_Int(2, 1, INT32_MAX));
+				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(
+								DEFAULT_RATE, 1, INT32_MAX),
+				SPA_FORMAT_AUDIO_channels, SPA_POD_CHOICE_RANGE_Int(
+								DEFAULT_CHANNELS, 1, INT32_MAX));
 		}
 		break;
 	default:
@@ -412,11 +402,10 @@ impl_node_port_enum_params(void *object, int seq,
 			SPA_PARAM_BUFFERS_buffers, SPA_POD_CHOICE_RANGE_Int(1, 1, MAX_BUFFERS),
 			SPA_PARAM_BUFFERS_blocks,  SPA_POD_Int(this->blocks),
 			SPA_PARAM_BUFFERS_size,    SPA_POD_CHOICE_RANGE_Int(
-								MAX_SAMPLES * this->stride,
+								this->quantum_limit * this->stride,
 								16 * this->stride,
 								INT32_MAX),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(this->stride),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(16));
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(this->stride));
 		break;
 	case SPA_PARAM_Meta:
 		switch (result.index) {
@@ -460,7 +449,7 @@ impl_node_port_enum_params(void *object, int seq,
 static int clear_buffers(struct impl *this, struct port *port)
 {
 	if (port->n_buffers > 0) {
-		spa_log_debug(this->log, NAME " %p: clear buffers %p", this, port);
+		spa_log_debug(this->log, "%p: clear buffers %p", this, port);
 		port->n_buffers = 0;
 		spa_list_init(&port->queue);
 	}
@@ -474,7 +463,7 @@ static int queue_buffer(struct impl *this, struct port *port, struct buffer *b)
 
 	spa_list_append(&port->queue, &b->link);
 	SPA_FLAG_SET(b->flags, BUFFER_FLAG_QUEUED);
-	spa_log_trace_fp(this->log, NAME " %p: queue buffer %d", this, b->id);
+	spa_log_trace_fp(this->log, "%p: queue buffer %d", this, b->id);
 	return 0;
 }
 
@@ -488,7 +477,7 @@ static struct buffer *dequeue_buffer(struct impl *this, struct port *port)
 	b = spa_list_first(&port->queue, struct buffer, link);
 	spa_list_remove(&b->link);
 	SPA_FLAG_CLEAR(b->flags, BUFFER_FLAG_QUEUED);
-	spa_log_trace_fp(this->log, NAME " %p: dequeue buffer %d", this, b->id);
+	spa_log_trace_fp(this->log, "%p: dequeue buffer %d", this, b->id);
 	return b;
 }
 
@@ -512,6 +501,10 @@ static int calc_width(struct spa_audio_info *info)
 	case SPA_AUDIO_FORMAT_S24_OE:
 	case SPA_AUDIO_FORMAT_U24:
 		return 3;
+	case SPA_AUDIO_FORMAT_F64P:
+	case SPA_AUDIO_FORMAT_F64:
+	case SPA_AUDIO_FORMAT_F64_OE:
+		return 8;
 	default:
 		return 4;
 	}
@@ -553,6 +546,10 @@ static int port_set_format(void *object,
 			if (memcmp(&info, &this->format, sizeof(struct spa_audio_info)))
 				return -EINVAL;
 		} else {
+			if (info.info.raw.format == 0 ||
+			    info.info.raw.channels == 0)
+				return -EINVAL;
+
 			this->ops.fmt = info.info.raw.format;
 			this->ops.n_channels = info.info.raw.channels;
 			this->ops.cpu_flags = this->cpu_flags;
@@ -575,7 +572,7 @@ static int port_set_format(void *object,
 		if (!port->have_format) {
 			this->n_formats++;
 			port->have_format = true;
-			spa_log_debug(this->log, NAME " %p: set format on port %d",
+			spa_log_debug(this->log, "%p: set format on port %d",
 					this, port_id);
 		}
 	}
@@ -625,17 +622,20 @@ impl_node_port_use_buffers(void *object,
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
-	spa_log_debug(this->log, NAME " %p: use %d buffers on port %d:%d",
+	spa_log_debug(this->log, "%p: use %d buffers on port %d:%d",
 			this, n_buffers, direction, port_id);
 
 	spa_return_val_if_fail(CHECK_PORT(this, direction, port_id), -EINVAL);
 
 	port = GET_PORT(this, direction, port_id);
 
-	spa_return_val_if_fail(port->have_format, -EIO);
-
 	clear_buffers(this, port);
 
+	if (n_buffers > 0 && !port->have_format)
+		return -EIO;
+	if (n_buffers > MAX_BUFFERS)
+		return -ENOSPC;
+
 	for (i = 0; i < n_buffers; i++) {
 		struct buffer *b;
 		struct spa_data *d = buffers[i]->datas;
@@ -645,19 +645,20 @@ impl_node_port_use_buffers(void *object,
 		b->flags = 0;
 		b->id = i;
 		b->h = spa_buffer_find_meta_data(buffers[i], SPA_META_Header, sizeof(*b->h));
+		b->buf = *buffers[i];
 
 		if (d[0].data == NULL) {
-			spa_log_error(this->log, NAME " %p: invalid memory on buffer %p", this,
+			spa_log_error(this->log, "%p: invalid memory on buffer %p", this,
 				      buffers[i]);
 			return -EINVAL;
 		}
-		if (!SPA_IS_ALIGNED(d[0].data, 16)) {
-			spa_log_warn(this->log, NAME " %p: memory on buffer %d not aligned", this, i);
+		if (!SPA_IS_ALIGNED(d[0].data, this->max_align)) {
+			spa_log_warn(this->log, "%p: memory on buffer %d not aligned", this, i);
 		}
 		if (direction == SPA_DIRECTION_OUTPUT)
 			queue_buffer(this, port, b);
 
-		spa_log_debug(this->log, NAME " %p: port %d:%d buffer:%d n_data:%d data:%p maxsize:%d",
+		spa_log_debug(this->log, "%p: port %d:%d buffer:%d n_data:%d data:%p maxsize:%d",
 				this, direction, port_id, i,
 				buffers[i]->n_datas, d[0].data, d[0].maxsize);
 	}
@@ -676,7 +677,7 @@ impl_node_port_set_io(void *object,
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
-	spa_log_debug(this->log, NAME " %p: port %d:%d io %d %p/%zd", this,
+	spa_log_debug(this->log, "%p: port %d:%d io %d %p/%zd", this,
 			direction, port_id, id, data, size);
 
 	spa_return_val_if_fail(CHECK_PORT(this, direction, port_id), -EINVAL);
@@ -713,18 +714,18 @@ static int impl_node_process(void *object)
 	struct impl *this = object;
 	struct port *outport;
 	struct spa_io_buffers *outio;
-	uint32_t n_samples, n_buffers, i, maxsize;
+	uint32_t n_buffers, i, maxsize;
 	struct buffer **buffers;
-        struct buffer *outb;
+	struct buffer *outb;
 	const void **datas;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
 	outport = GET_OUT_PORT(this, 0);
-	outio = outport->io;
-	spa_return_val_if_fail(outio != NULL, -EIO);
+	if ((outio = outport->io) == NULL)
+		return -EIO;
 
-	spa_log_trace_fp(this->log, NAME " %p: status %p %d %d",
+	spa_log_trace_fp(this->log, "%p: status %p %d %d",
 			this, outio, outio->status, outio->buffer_id);
 
 	if (SPA_UNLIKELY(outio->status == SPA_STATUS_HAVE_DATA))
@@ -740,20 +741,22 @@ static int impl_node_process(void *object)
         datas = alloca(MAX_PORTS * sizeof(void *));
         n_buffers = 0;
 
-	maxsize = MAX_SAMPLES * this->stride;
+	maxsize = UINT32_MAX;
 
 	for (i = 0; i < this->last_port; i++) {
 		struct port *inport = GET_IN_PORT(this, i);
 		struct spa_io_buffers *inio = NULL;
 		struct buffer *inb;
+		struct spa_data *bd;
+		uint32_t size, offs;
 
-		if (SPA_UNLIKELY(!inport->valid ||
+		if (SPA_UNLIKELY(!PORT_VALID(inport) ||
 		    (inio = inport->io) == NULL ||
 		    inio->buffer_id >= inport->n_buffers ||
 		    inio->status != SPA_STATUS_HAVE_DATA)) {
-			spa_log_trace_fp(this->log, NAME " %p: skip input idx:%d valid:%d "
+			spa_log_trace_fp(this->log, "%p: skip input idx:%d valid:%d "
 					"io:%p status:%d buf_id:%d n_buffers:%d", this,
-				i, inport->valid, inio,
+				i, PORT_VALID(inport), inio,
 				inio ? inio->status : -1,
 				inio ? inio->buffer_id : SPA_ID_INVALID,
 				inport->n_buffers);
@@ -761,38 +764,46 @@ static int impl_node_process(void *object)
 		}
 
 		inb = &inport->buffers[inio->buffer_id];
-		maxsize = SPA_MIN(inb->buffer->datas[0].chunk->size, maxsize);
+		bd = &inb->buffer->datas[0];
 
-		spa_log_trace_fp(this->log, NAME " %p: mix input %d %p->%p %d %d %d", this,
-				i, inio, outio, inio->status, inio->buffer_id, maxsize);
+		offs = SPA_MIN(bd->chunk->offset, bd->maxsize);
+		size = SPA_MIN(bd->maxsize - offs, bd->chunk->size);
+		maxsize = SPA_MIN(size, maxsize);
 
-		datas[n_buffers] = inb->buffer->datas[0].data;
-		buffers[n_buffers++] = inb;
+		spa_log_trace_fp(this->log, "%p: mix input %d %p->%p %d %d %d:%d/%d", this,
+				i, inio, outio, inio->status, inio->buffer_id,
+				offs, size, this->stride);
+
+		if (!SPA_FLAG_IS_SET(bd->chunk->flags, SPA_CHUNK_FLAG_EMPTY)) {
+			datas[n_buffers] = SPA_PTROFF(bd->data, offs, void);
+			buffers[n_buffers++] = inb;
+		}
 		inio->status = SPA_STATUS_NEED_DATA;
 	}
 
 	outb = dequeue_buffer(this, outport);
         if (SPA_UNLIKELY(outb == NULL)) {
-                spa_log_trace(this->log, NAME " %p: out of buffers", this);
+                spa_log_trace(this->log, "%p: out of buffers (%d)", this,
+				outport->n_buffers);
                 return -EPIPE;
         }
 
-	n_samples = maxsize / this->stride;
-
 	if (n_buffers == 1) {
 		*outb->buffer = *buffers[0]->buffer;
-	}
-	else {
-		outb->buffer->n_datas = 1;
-		outb->buffer->datas = outb->datas;
-		outb->datas[0].data = this->empty;
-		outb->datas[0].chunk = outb->chunk;
-		outb->datas[0].chunk->offset = 0;
-		outb->datas[0].chunk->size = n_samples * this->stride;
-		outb->datas[0].chunk->stride = this->stride;
-		outb->datas[0].maxsize = maxsize;
+	} else {
+		struct spa_data *d = outb->buf.datas;
 
-		mix_ops_process(&this->ops, outb->datas[0].data, datas, n_buffers, n_samples);
+		*outb->buffer = outb->buf;
+
+		maxsize = SPA_MIN(maxsize, d[0].maxsize);
+
+		d[0].chunk->offset = 0;
+		d[0].chunk->size = maxsize;
+		d[0].chunk->stride = this->stride;
+		SPA_FLAG_UPDATE(d[0].chunk->flags, SPA_CHUNK_FLAG_EMPTY, n_buffers == 0);
+
+		mix_ops_process(&this->ops, d[0].data,
+				datas, n_buffers, maxsize / this->stride);
 	}
 
 	outio->buffer_id = outb->id;
@@ -839,11 +850,14 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 static int impl_clear(struct spa_handle *handle)
 {
 	struct impl *this;
+	uint32_t i;
 
 	spa_return_val_if_fail(handle != NULL, -EINVAL);
 
 	this = (struct impl *) handle;
 
+	for (i = 0; i < MAX_PORTS; i++)
+		free(this->in_ports[i]);
 	mix_ops_free(&this->ops);
 	return 0;
 }
@@ -864,6 +878,7 @@ impl_init(const struct spa_handle_factory *factory,
 {
 	struct impl *this;
 	struct port *port;
+	uint32_t i;
 
 	spa_return_val_if_fail(factory != NULL, -EINVAL);
 	spa_return_val_if_fail(handle != NULL, -EINVAL);
@@ -874,10 +889,20 @@ impl_init(const struct spa_handle_factory *factory,
 	this = (struct impl *) handle;
 
 	this->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
-	this->cpu = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_CPU);
+	spa_log_topic_init(this->log, log_topic);
 
-	if (this->cpu)
+	this->cpu = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_CPU);
+	if (this->cpu) {
 		this->cpu_flags = spa_cpu_get_flags(this->cpu);
+		this->max_align = SPA_MIN(MAX_ALIGN, spa_cpu_get_max_align(this->cpu));
+	}
+
+	for (i = 0; info && i < info->n_items; i++) {
+		const char *k = info->items[i].key;
+		const char *s = info->items[i].value;
+		if (spa_streq(k, "clock.quantum-limit"))
+			spa_atou32(s, &this->quantum_limit, 0);
+	}
 
 	spa_hook_list_init(&this->hooks);
 
