@@ -58,6 +58,10 @@ struct props {
 #define BUFFER_SIZE	(8192*8)
 #define RATE_CTL_DIFF_MAX 0.005
 
+/* Wait for two cycles before trying to sync ISO. On start/driver reassign,
+ * first cycle may have strange number of samples. */
+#define RESYNC_CYCLES 2
+
 struct buffer {
 	uint32_t id;
 #define BUFFER_FLAG_OUT	(1<<0)
@@ -129,8 +133,10 @@ struct impl {
 	unsigned int following:1;
 	unsigned int is_output:1;
 	unsigned int flush_pending:1;
+	unsigned int iso_pending:1;
 
 	unsigned int is_duplex:1;
+	unsigned int is_internal:1;
 
 	struct spa_source source;
 	int timerfd;
@@ -159,8 +165,7 @@ struct impl {
 
 	int need_flush;
 	bool fragment;
-	bool resync;
-	bool have_iso_packet;
+	uint32_t resync;
 	uint32_t block_size;
 	uint8_t buffer[BUFFER_SIZE];
 	uint32_t buffer_used;
@@ -307,7 +312,7 @@ static int do_reassign_io(struct spa_loop *loop,
 	bool following;
 
 	if (this->position != info->position || this->clock != info->clock)
-		this->resync = true;
+		this->resync = RESYNC_CYCLES;
 
 	this->position = info->position;
 	this->clock = info->clock;
@@ -460,6 +465,7 @@ static uint64_t get_reference_time(struct impl *this, uint64_t *duration_ns_ret)
 {
 	struct port *port = &this->port;
 	uint64_t t, duration_ns;
+	bool resampling;
 
 	if (!this->process_rate || !this->process_duration) {
 		if (this->position) {
@@ -481,9 +487,12 @@ static uint64_t get_reference_time(struct impl *this, uint64_t *duration_ns_ret)
 			/ port->current_format.info.raw.rate);
 
 	/* Account for resampling delay */
-	if (port->rate_match && this->clock && SPA_FLAG_IS_SET(port->rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE))
+	resampling = (port->current_format.info.raw.rate != this->process_rate) || this->following;
+	if (port->rate_match && this->clock && resampling) {
 		t -= (uint64_t)port->rate_match->delay * SPA_NSEC_PER_SEC
 			/ this->clock->rate.denom;
+		t += SPA_NSEC_PER_SEC / port->current_format.info.raw.rate;
+	}
 
 	return t;
 }
@@ -715,6 +724,9 @@ static int flush_data(struct impl *this, uint64_t now_time)
 	if (!this->flush_timer_source.loop && !this->transport->iso_io)
 		return -EIO;
 
+	if (this->transport->iso_io && !this->iso_pending)
+		return 0;
+
 	total_frames = 0;
 again:
 	written = 0;
@@ -787,7 +799,7 @@ again:
 	if (this->transport->iso_io) {
 		struct spa_bt_iso_io *iso_io = this->transport->iso_io;
 
-		if (this->need_flush && !this->have_iso_packet) {
+		if (this->need_flush) {
 			size_t avail = SPA_MIN(this->buffer_used, sizeof(iso_io->buf));
 
 			spa_log_trace(this->log, "%p: ISO put fd:%d size:%u sn:%u ts:%u now:%"PRIu64,
@@ -798,7 +810,7 @@ again:
 			memcpy(iso_io->buf, this->buffer, avail);
 			iso_io->size = avail;
 			iso_io->timestamp = this->timestamp;
-			this->have_iso_packet = true;
+			this->iso_pending = false;
 
 			reset_buffer(this);
 		}
@@ -953,9 +965,7 @@ static void media_iso_pull(struct spa_bt_iso_io *iso_io)
 	struct port *port = &this->port;
 	const double period = 0.1 * SPA_NSEC_PER_SEC;
 	uint64_t duration_ns;
-	double value, target, err;
-
-	this->have_iso_packet = false;
+	double value, target, err, max_err;
 
 	if (this->resync || !this->position) {
 		spa_bt_rate_control_init(&port->ratectl, 0);
@@ -978,19 +988,28 @@ static void media_iso_pull(struct spa_bt_iso_io *iso_io)
 	value = (int64_t)iso_io->now - (int64_t)get_reference_time(this, &duration_ns);
 	target = iso_io->duration * 3/2;
 	err = value - target;
+	max_err = iso_io->duration;
 
-	if (err > iso_io->duration) {
-		uint32_t req = err * port->current_format.info.raw.rate / SPA_NSEC_PER_SEC;
+	if (err > max_err || (iso_io->resync && err > 0)) {
+		unsigned int req = err * port->current_format.info.raw.rate / SPA_NSEC_PER_SEC;
 
-		spa_log_debug(this->log, "%p: ISO sync reset frames:%u", this, (unsigned int)req);
+		if (req > 0) {
+			spa_bt_rate_control_init(&port->ratectl, 0);
+			drop_frames(this, req);
+			spa_log_debug(this->log, "%p: ISO sync skip frames:%u resync:%d",
+					this, req, iso_io->resync);
+		}
+	} else if (-err > max_err || (iso_io->resync && -err > 0)) {
+		unsigned int req = -err * port->current_format.info.raw.rate / SPA_NSEC_PER_SEC;
+		static const uint8_t empty[8192] = {0};
 
-		spa_bt_rate_control_init(&port->ratectl, 0);
-		drop_frames(this, req);
-	} else if (-err > iso_io->duration) {
-		uint32_t req = -err * port->current_format.info.raw.rate / SPA_NSEC_PER_SEC;
-
-		spa_log_debug(this->log, "%p: ISO sync skip flush frames:%u", this, (unsigned int)req);
-		return;
+		if (req > 0) {
+			spa_bt_rate_control_init(&port->ratectl, 0);
+			req = SPA_MIN(req, sizeof(empty) / port->frame_size);
+			add_data(this, empty, req * port->frame_size);
+			spa_log_debug(this->log, "%p: ISO sync pad frames:%u resync:%d",
+					this, req, iso_io->resync);
+		}
 	} else {
 		spa_bt_rate_control_update(&port->ratectl, err, 0,
 				iso_io->duration, period, RATE_CTL_DIFF_MAX);
@@ -1002,7 +1021,10 @@ static void media_iso_pull(struct spa_bt_iso_io *iso_io)
 				port->ratectl.corr);
 	}
 
+	iso_io->resync = false;
+
 done:
+	this->iso_pending = true;
 	flush_data(this, this->current_time);
 }
 
@@ -1215,9 +1237,9 @@ static int transport_start(struct impl *this)
 	this->flush_source.rmask = 0;
 	spa_loop_add_source(this->data_loop, &this->flush_source);
 
-	this->resync = true;
-
+	this->resync = RESYNC_CYCLES;
 	this->flush_pending = false;
+	this->iso_pending = false;
 
 	this->transport_started = true;
 
@@ -1375,7 +1397,8 @@ static void emit_node_info(struct impl *this, bool full)
 {
 	struct spa_dict_item node_info_items[] = {
 		{ SPA_KEY_DEVICE_API, "bluez5" },
-		{ SPA_KEY_MEDIA_CLASS, this->is_output ? "Audio/Sink" : "Stream/Input/Audio" },
+		{ SPA_KEY_MEDIA_CLASS, this->is_internal ? "Audio/Sink/Internal" :
+		  this->is_output ? "Audio/Sink" : "Stream/Input/Audio" },
 		{ "media.name", ((this->transport && this->transport->device->name) ?
 					this->transport->device->name : this->codec->bap ? "BAP" : "A2DP" ) },
 		{ SPA_KEY_NODE_DRIVER, this->is_output ? "true" : "false" },
@@ -1796,6 +1819,8 @@ static int impl_node_process(void *object)
 
 	if (io->status == SPA_STATUS_HAVE_DATA && io->buffer_id < port->n_buffers) {
 		struct buffer *b = &port->buffers[io->buffer_id];
+		struct spa_data *d = b->buf->datas;
+		unsigned int frames;
 
 		if (!SPA_FLAG_IS_SET(b->flags, BUFFER_FLAG_OUT)) {
 			spa_log_warn(this->log, "%p: buffer %u in use", this, io->buffer_id);
@@ -1803,7 +1828,8 @@ static int impl_node_process(void *object)
 			return -EINVAL;
 		}
 
-		spa_log_trace(this->log, "%p: queue buffer %u", this, io->buffer_id);
+		frames = d ? d[0].chunk->size / port->frame_size : 0;
+		spa_log_trace(this->log, "%p: queue buffer %u frames:%u", this, io->buffer_id, frames);
 
 		spa_list_append(&port->ready, &b->link);
 		SPA_FLAG_CLEAR(b->flags, BUFFER_FLAG_OUT);
@@ -1831,7 +1857,8 @@ static int impl_node_process(void *object)
 	}
 
 	this->process_time = this->current_time;
-	this->resync = false;
+	if (this->resync)
+		--this->resync;
 
 	setup_matching(this);
 
@@ -2056,6 +2083,9 @@ impl_init(const struct spa_handle_factory *factory,
 	if (info && (str = spa_dict_lookup(info, "api.bluez5.a2dp-duplex")) != NULL)
 		this->is_duplex = spa_atob(str);
 
+	if (info && (str = spa_dict_lookup(info, "api.bluez5.internal")) != NULL)
+		this->is_internal = spa_atob(str);
+
 	if (info && (str = spa_dict_lookup(info, SPA_KEY_API_BLUEZ5_TRANSPORT)))
 		sscanf(str, "pointer:%p", &this->transport);
 
