@@ -1,26 +1,6 @@
-/* PipeWire
- *
- * Copyright © 2021 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2021 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <string.h>
 #include <stdio.h>
@@ -41,7 +21,7 @@
 #include <spa/utils/json.h>
 #include <spa/utils/ringbuffer.h>
 #include <spa/utils/dll.h>
-#include <spa/debug/pod.h>
+#include <spa/debug/types.h>
 #include <spa/pod/builder.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/param/latency-utils.h>
@@ -52,6 +32,7 @@
 #include <pipewire/private.h>
 
 #include <pulse/pulseaudio.h>
+#include "module-protocol-pulse/defs.h"
 #include "module-protocol-pulse/format.h"
 
 /** \page page_module_pulse_tunnel PipeWire Module: Pulse Tunnel
@@ -125,19 +106,19 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 #define DEFAULT_CHANNELS 2
 #define DEFAULT_POSITION "[ FL FR ]"
 
-#define MODULE_USAGE	"[ remote.name=<remote> ] "				\
-			"[ node.latency=<latency as fraction> ] "		\
-			"[ node.name=<name of the nodes> ] "			\
-			"[ node.description=<description of the nodes> ] "	\
-			"[ node.target=<remote node target name or serial> ] "	\
-			"[ audio.format=<sample format> ] "			\
-			"[ audio.rate=<sample rate> ] "				\
-			"[ audio.channels=<number of channels> ] "		\
-			"[ audio.position=<channel map> ] "			\
+#define MODULE_USAGE	"( remote.name=<remote> ] "				\
+			"( node.latency=<latency as fraction> ] "		\
+			"( node.name=<name of the nodes> ] "			\
+			"( node.description=<description of the nodes> ] "	\
+			"( node.target=<remote node target name or serial> ] "	\
+			"( audio.format=<sample format> ] "			\
+			"( audio.rate=<sample rate> ] "				\
+			"( audio.channels=<number of channels> ] "		\
+			"( audio.position=<channel map> ] "			\
 			"pulse.server.address=<address> "			\
-			"pulse.latency=<latency in msec> "			\
-			"[ tunnel.mode=source|sink "				\
-			"[ stream.props=<properties> ] "
+			"( pulse.latency=<latency in msec, default 200> ) "	\
+			"( tunnel.mode=source|sink, default sink ) "				\
+			"( stream.props=<properties> ) "
 
 
 static const struct spa_dict_item module_props[] = {
@@ -154,6 +135,7 @@ static const struct spa_dict_item module_props[] = {
 
 struct impl {
 	struct pw_context *context;
+	struct pw_loop *main_loop;
 
 #define MODE_SINK	0
 #define MODE_SOURCE	1
@@ -251,23 +233,22 @@ static void stream_state_changed(void *d, enum pw_stream_state old,
 	}
 }
 
-static void update_rate(struct impl *impl, bool playback)
+static void update_rate(struct impl *impl, uint32_t filled)
 {
 	float error, corr;
+	uint32_t current_latency;
 
 	if (impl->rate_match == NULL)
 		return;
 
-	if (playback)
-		error = (float)impl->target_latency - (float)impl->current_latency;
-	else
-		error = (float)impl->current_latency - (float)impl->target_latency;
+	current_latency = impl->current_latency + filled;
+	error = (float)impl->target_latency - (float)(current_latency);
 	error = SPA_CLAMP(error, -impl->max_error, impl->max_error);
 
 	corr = spa_dll_update(&impl->dll, error);
 	pw_log_debug("error:%f corr:%f current:%u target:%u",
 			error, corr,
-			impl->current_latency, impl->target_latency);
+			current_latency, impl->target_latency);
 
 	SPA_FLAG_SET(impl->rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE);
 	impl->rate_match->rate = 1.0f / corr;
@@ -302,7 +283,7 @@ static void playback_stream_process(void *d)
                                         size, RINGBUFFER_SIZE);
 		impl->resync = true;
 	} else {
-		update_rate(impl, true);
+		update_rate(impl, filled / impl->frame_size);
 	}
 	spa_ringbuffer_write_data(&impl->ring,
 				impl->buffer, RINGBUFFER_SIZE,
@@ -343,7 +324,7 @@ static void capture_stream_process(void *d)
 			avail = impl->target_buffer;
 			index += avail - impl->target_buffer;
 		} else {
-			update_rate(impl, false);
+			update_rate(impl, avail / impl->frame_size);
 		}
 		spa_ringbuffer_read_data(&impl->ring,
 				impl->buffer, RINGBUFFER_SIZE,
@@ -435,6 +416,20 @@ static int create_stream(struct impl *impl)
 	return 0;
 }
 
+static int
+do_schedule_destroy(struct spa_loop *loop,
+	bool async, uint32_t seq, const void *data, size_t size, void *user_data)
+{
+	struct impl *impl = user_data;
+	pw_impl_module_schedule_destroy(impl->module);
+	return 0;
+}
+
+void module_schedule_destroy(struct impl *impl)
+{
+	pw_loop_invoke(impl->main_loop, do_schedule_destroy, 1, NULL, 0, false, impl);
+}
+
 static void context_state_cb(pa_context *c, void *userdata)
 {
 	struct impl *impl = userdata;
@@ -456,7 +451,7 @@ static void context_state_cb(pa_context *c, void *userdata)
 		break;
 	}
 	if (do_destroy)
-		pw_impl_module_schedule_destroy(impl->module);
+		module_schedule_destroy(impl);
 }
 
 static void stream_state_cb(pa_stream *s, void * userdata)
@@ -478,7 +473,7 @@ static void stream_state_cb(pa_stream *s, void * userdata)
 		break;
 	}
 	if (do_destroy)
-		pw_impl_module_schedule_destroy(impl->module);
+		module_schedule_destroy(impl);
 }
 
 static void stream_read_request_cb(pa_stream *s, size_t length, void *userdata)
@@ -531,7 +526,6 @@ static void stream_read_request_cb(pa_stream *s, size_t length, void *userdata)
 
 	pa_stream_get_latency(impl->pa_stream, &latency, &negative);
 	impl->current_latency = latency * impl->info.rate / SPA_USEC_PER_SEC;
-	impl->current_latency += filled / impl->frame_size;
 
 	spa_ringbuffer_write_update(&impl->ring, index);
 }
@@ -556,7 +550,6 @@ static void stream_write_request_cb(pa_stream *s, size_t length, void *userdata)
 
 	pa_stream_get_latency(impl->pa_stream, &latency, &negative);
 	impl->current_latency = latency * impl->info.rate / SPA_USEC_PER_SEC;
-	impl->current_latency += avail / impl->frame_size;
 
 	while (avail < (int32_t)length) {
 		uint32_t maxsize = SPA_ROUND_DOWN(sizeof(impl->empty), impl->frame_size);
@@ -773,7 +766,7 @@ error_unlock:
 	pa_threaded_mainloop_unlock(impl->pa_mainloop);
 error:
 	pw_log_error("failed to connect: %s", pa_strerror(res));
-	return -res;
+	return err_to_res(res);
 }
 
 
@@ -980,6 +973,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 
 	impl->module = module;
 	impl->context = context;
+	impl->main_loop = pw_context_get_main_loop(context);
 
 	spa_ringbuffer_init(&impl->ring);
 	impl->buffer = calloc(1, RINGBUFFER_SIZE);
